import logging
from enum import Enum

from django.utils.translation import gettext as _

from reversion import create_revision, set_comment
from tqdm import tqdm

from aleksis.apps.kolego import models as kolego_models
from aleksis.apps.lesrooster.models import ValidityRange

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    date_to_untis_date,
    get_first_period,
    get_last_period,
    move_weekday_to_range,
    run_default_filter,
    untis_date_to_date,
)

logger = logging.getLogger(__name__)
unknown_reason, __ = kolego_models.AbsenceReason.objects.get_or_create(short_name="?")


class AbsenceType(Enum):
    GROUP = 100
    TEACHER = 101
    ROOM = 102


def import_absences(
    validity_range: ValidityRange,
    absence_reasons_ref,
    slots_ref,
    teachers_ref,
    classes_ref,
    rooms_ref,
):
    ref = {}

    untis_term_start = date_to_untis_date(validity_range.date_start)
    untis_term_end = date_to_untis_date(validity_range.date_end)

    created_substitutions = []

    # Get absences
    absences = (
        run_default_filter(validity_range, mysql_models.Absence.objects, filter_term=False)
        .filter(datefrom__lte=untis_term_end, dateto__gte=untis_term_start)
        .order_by("absence_id")
    )

    existing_absences = []
    for absence in tqdm(absences, desc="Import absences", **TQDM_DEFAULTS):
        import_ref = absence.absence_id

        logger.info(f"Import absence {import_ref}")

        if absence.absence_reason_id == 0:
            reason = unknown_reason
        else:
            reason = absence_reasons_ref[absence.absence_reason_id]

        # Build values
        type_ = absence.typea
        date_from = untis_date_to_date(absence.datefrom)
        date_to = untis_date_to_date(absence.dateto)
        period_from = absence.lessonfrom
        period_to = absence.lessonto
        weekday_from = date_from.weekday()
        weekday_to = date_to.weekday()

        # Check min/max weekdays
        weekday_from = move_weekday_to_range(slots_ref, weekday_from)
        weekday_to = move_weekday_to_range(slots_ref, weekday_to)

        # Check min/max periods
        first_period = get_first_period(slots_ref, weekday_from)
        last_period = get_last_period(slots_ref, weekday_from)

        if period_from == 0:
            period_from = first_period
        if period_to == 0:
            period_to = last_period

        slot_from = slots_ref[weekday_from][period_from]
        slot_to = slots_ref[weekday_to][period_to]
        comment = absence.text if absence.text else ""

        datetime_start = slot_from.get_datetime_start(date_from)
        datetime_end = slot_to.get_datetime_end(date_to)

        group = None
        teacher = None
        room = None

        if type_ == AbsenceType.GROUP.value:
            group = classes_ref[absence.ida]  # noqa
        elif type_ == AbsenceType.TEACHER.value:
            teacher = teachers_ref[absence.ida]
        elif type_ == AbsenceType.ROOM.value:
            room = rooms_ref[absence.ida]  # noqa

        if teacher:
            defaults = {
                "reason": reason,
                "person": teacher,
                "datetime_start": datetime_start,
                "datetime_end": datetime_end,
                "comment": comment,
            }

            new_absence, created = kolego_models.Absence.objects.update_or_create(
                extended_data__import_ref_untis=import_ref, defaults=defaults
            )

            new_absence.extended_data["import_ref_untis"] = import_ref
            new_absence.save()

            if created:
                logger.info("  New absence created")

            existing_absences.append(import_ref)
            ref[import_ref] = new_absence

    # Delete all no longer existing absences
    for a in kolego_models.Absence.objects.filter(
        extended_data__import_ref_untis__isnull=False,
        datetime_start__date__lte=validity_range.date_end,
        datetime_end__date__gte=validity_range.date_start,
    ):
        if a.extended_data["import_ref_untis"] not in existing_absences:
            logger.info(f"Absence {a.id} deleted")
            with create_revision():
                set_comment(_("Deleted by Untis import"))
                a.delete()

    return ref, created_substitutions
