import logging
import re
from datetime import datetime, time, timedelta
from enum import Enum
from typing import Dict, Tuple

from django.utils.translation import gettext as _

import recurrence
from reversion import create_revision, set_comment
from tqdm import tqdm

from aleksis.apps.cursus import models as cursus_models
from aleksis.apps.kolego import models as kolego_models
from aleksis.apps.lesrooster import models as lesrooster_models
from aleksis.apps.lesrooster.models import ValidityRange
from aleksis.core import models as core_models
from aleksis.core.util.core_helpers import get_site_preferences

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    connect_untis_fields,
    run_default_filter,
    untis_colour_to_hex,
    untis_split_first,
    untis_split_second,
)

logger = logging.getLogger(__name__)


class CommonDataId(Enum):
    PERIOD = 40


def import_subjects(validity_range: ValidityRange) -> Dict[int, cursus_models.Subject]:
    """Import subjects."""
    subjects_ref = {}

    # Get subjects
    subjects = run_default_filter(validity_range, mysql_models.Subjects.objects, filter_term=False)

    for subject in tqdm(subjects, desc="Import subjects", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not subject.name:
            raise RuntimeError(
                f"Subject ID {subject.subject_id}: Cannot import subject without short name."
            )

        # Build values
        short_name = subject.name
        name = subject.longname if subject.longname else short_name
        colour_fg = untis_colour_to_hex(subject.forecolor)
        colour_bg = untis_colour_to_hex(subject.backcolor)
        import_ref = subject.subject_id

        logger.info(f"Import subject {short_name} …")

        # Get or create subject object by short name
        new_subject, created = cursus_models.Subject.objects.get_or_create(
            short_name=short_name,
            defaults={
                "name": name,
                "colour_fg": colour_fg,
                "colour_bg": colour_bg,
            },
        )

        if created:
            logger.info("  New subject created")

        # Force sync
        changed = False
        if get_site_preferences()["untis_mysql__update_subjects"] and (
            new_subject.name != name
            or new_subject.colour_fg != colour_fg
            or new_subject.colour_bg != colour_bg
        ):
            new_subject.name = name
            new_subject.colour_fg = untis_colour_to_hex(subject.forecolor)
            new_subject.colour_bg = untis_colour_to_hex(subject.backcolor)
            changed = True

            logger.info("  Name, foreground and background colour updated")

        if new_subject.extended_data.get("import_ref_untis") != import_ref:
            new_subject.extended_data["import_ref_untis"] = import_ref
            changed = True

            logger.info("  Import reference updated")

        if changed:
            new_subject.save()

        subjects_ref[import_ref] = new_subject

    return subjects_ref


def import_teachers(validity_range: ValidityRange) -> Dict[int, core_models.Person]:
    """Import teachers."""
    teachers_ref = {}

    # Get teachers
    teachers = run_default_filter(validity_range, mysql_models.Teacher.objects)

    for teacher in tqdm(teachers, desc="Import teachers", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not teacher.name:
            raise RuntimeError(
                f"Teacher ID {teacher.teacher_id}: Cannot import teacher without short name."
            )

        # Build values
        short_name = teacher.name
        first_name = teacher.firstname if teacher.firstname else "?"
        last_name = teacher.longname if teacher.longname else teacher.name
        import_ref = teacher.teacher_id

        logger.info(f"Import teacher {short_name} (as person) …")

        try:
            new_teacher = core_models.Person.objects.get(short_name__iexact=short_name)
        except core_models.Person.DoesNotExist:
            new_teacher = core_models.Person.objects.create(
                short_name=short_name,
                first_name=first_name,
                last_name=last_name,
            )
            logger.info("  New person created")

        changed = False
        if get_site_preferences()["untis_mysql__update_persons_name"] and (
            new_teacher.first_name != first_name or new_teacher.last_name != last_name
        ):
            new_teacher.first_name = first_name
            new_teacher.last_name = last_name
            changed = True
            logger.info("  First and last name updated")

        if (
            get_site_preferences()["untis_mysql__update_persons_short_name"]
            and new_teacher.short_name != short_name
        ):
            new_teacher.short_name = short_name
            changed = True
            logger.info("  Short name updated")

        if new_teacher.extended_data.get("import_ref_untis") != import_ref:
            new_teacher.extended_data["import_ref_untis"] = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_teacher.save()

        teachers_ref[teacher.teacher_id] = new_teacher

    return teachers_ref


def import_classes(
    validity_range: ValidityRange, teachers_ref: Dict[int, core_models.Person]
) -> Dict[int, core_models.Group]:
    """Import classes."""
    classes_ref = {}

    group_type = get_site_preferences()["untis_mysql__group_type_class_groups"]

    # Get classes
    course_classes = run_default_filter(
        validity_range, mysql_models.Class.objects, filter_term=True
    )

    for class_ in tqdm(course_classes, desc="Import classes", **TQDM_DEFAULTS):
        # Check if needed data are provided
        if not class_.name:
            raise RuntimeError(
                f"Class ID {class_.teacher_id}: Cannot import class without short name."
            )

        # Build values
        short_name = class_.name
        name = class_.longname if class_.longname else short_name
        teacher_ids = untis_split_first(class_.teacherids, int) if class_.teacherids else []
        owners = [teachers_ref[t_id] for t_id in teacher_ids]
        import_ref = class_.class_id

        logger.info(f"Import class {short_name} (as group) …")

        try:
            new_group = core_models.Group.objects.get(
                short_name__iexact=short_name,
                school_term__in=[None, validity_range.school_term],
            )
        except core_models.Group.DoesNotExist:
            # Determine collisions of long name
            while core_models.Group.objects.filter(
                name=name, school_term=validity_range.school_term
            ).exists():
                if not get_site_preferences()["untis_mysql__disambiguate_groups_name"]:
                    # Do not try to disambiguate it not enabled
                    break

                suffix = f" ({short_name})"
                if name.endswith(suffix):
                    # First try, add a counter
                    name = f"{name} (2)"
                else:
                    # Second or more tries, determine last counter, if any
                    match = re.match(r"^(.*)(\(\d+\))$", name)
                    if match:
                        # Counter found, increase
                        prefix = match.group(1)
                        counter = int(match.group(2)) + 1
                        name = f"{prefix}({counter})"
                    else:
                        # Counter not found, add one
                        name = f"{name} (2)"

            new_group = core_models.Group.objects.create(
                short_name=short_name,
                name=name,
                school_term=validity_range.school_term,
            )
            logger.info("  New group created")

        changed = False

        if (
            get_site_preferences()["untis_mysql__update_groups_short_name"]
            and new_group.short_name != short_name
        ):
            new_group.short_name = short_name
            changed = True
            logger.info("  Short name updated")

        if get_site_preferences()["untis_mysql__update_groups_name"] and new_group.name != name:
            new_group.name = name
            changed = True
            logger.info("  Name updated")

        if new_group.extended_data.get("import_ref_untis") != import_ref:
            new_group.extended_data["import_ref_untis"] = import_ref
            changed = True
            logger.info("  Import reference updated")

        if new_group.school_term != validity_range.school_term:
            new_group.school_term = validity_range.school_term
            changed = True
            logger.info("  School term updated")

        if group_type and new_group.group_type != group_type:
            new_group.group_type = group_type
            changed = True
            logger.info("  Group type updated")

        if changed:
            new_group.save()

        if get_site_preferences()["untis_mysql__overwrite_group_owners"]:
            new_group.owners.set(owners)
            logger.info("  Group owners set")
        else:
            new_group.owners.add(*owners)
            logger.info("  Group owners added")

        classes_ref[class_.class_id] = new_group

    return classes_ref


def import_rooms(validity_range: ValidityRange) -> Dict[int, core_models.Room]:
    """Import rooms."""
    ref = {}

    # Get rooms
    rooms = run_default_filter(validity_range, mysql_models.Room.objects)

    for room in tqdm(rooms, desc="Import rooms", **TQDM_DEFAULTS):
        if not room.name:
            raise RuntimeError(f"Room ID {room.room_id}: Cannot import room without short name.")

        # Build values
        short_name = room.name
        name = room.longname if room.longname else short_name
        import_ref = room.room_id

        logger.info(f"Import room {short_name} …")

        new_room, created = core_models.Room.objects.get_or_create(
            short_name=short_name,
            defaults={"name": name},
        )

        if created:
            logger.info("  New room created")

        changed = False

        if get_site_preferences()["untis_mysql__update_rooms_name"] and new_room.name != name:
            new_room.name = name
            changed = True
            logger.info("  Name updated")

        if new_room.extended_data.get("import_ref_untis") != import_ref:
            new_room.extended_data["import_ref_untis"] = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_room.save()

        ref[import_ref] = new_room

    return ref


def import_supervision_areas(
    validity_range: ValidityRange, breaks_ref, teachers_ref
) -> Dict[int, core_models.Room]:
    """Import supervision areas."""
    ref = {}

    # Get supervision areas
    areas = run_default_filter(validity_range, mysql_models.Corridor.objects, filter_term=False)

    for area in tqdm(areas, desc="Import supervision areas", **TQDM_DEFAULTS):
        if not area.name:
            raise RuntimeError(
                "Supervision area ID {}: Cannot import supervision area without short name.".format(
                    area.corridor_id
                )
            )

        short_name = area.name
        name = area.longname if area.longname else short_name
        import_ref = area.corridor_id

        logger.info(f"Import supervision area {short_name} …")

        new_area, created = core_models.Room.objects.get_or_create(
            short_name=short_name,
            defaults={
                "name": name,
            },
        )

        if created:
            logger.info("  New room created")

        changed = False

        if get_site_preferences()["untis_mysql__update_supervision_areas"] and (
            new_area.name != new_area.name
        ):
            new_area.name = name
            changed = True

            logger.info("  Name updated")

        if new_area.extended_data.get("import_ref_untis") != import_ref:
            new_area.extended_data["import_ref_untis"] = import_ref
            changed = True
            logger.info("  Import reference updated")

        if changed:
            new_area.save()

        logger.info("  Import supervisions for this area")

        # Parse raw data
        raw_supervisions = connect_untis_fields(area, "breaksupervision", 16)

        supervisions_ref = {}
        for raw_supervision in raw_supervisions:
            # Split more and get teacher id
            raw_supervision_2 = untis_split_second(raw_supervision, remove_empty=False)
            teacher_id = int(raw_supervision_2[1])
            term_id = int(raw_supervision_2[0])

            if term_id != validity_range.extended_data["import_ref_untis"]:
                continue

            if teacher_id in teachers_ref:
                # Get weekday, period after break and teacher
                weekday = int(raw_supervision_2[2]) - 1
                before_period_number = int(raw_supervision_2[3])
                teacher = teachers_ref[teacher_id]
                slot = breaks_ref[weekday][before_period_number]

                logger.info(
                    "Import supervision on weekday {} before the {}. period (teacher {})".format(
                        weekday, before_period_number, teacher
                    )
                )

                # Get or create
                possible_supervisions = lesrooster_models.Supervision.objects.filter(
                    break_slot=slot,
                    rooms__in=[new_area],
                )
                if possible_supervisions.exists():
                    new_supervision = possible_supervisions[0]
                    created = False
                else:
                    new_supervision = lesrooster_models.Supervision(
                        break_slot=slot,
                    )
                    created = True

                new_supervision.recurrence = new_supervision.build_recurrence(recurrence.WEEKLY)

                new_supervision.save()

                new_supervision.rooms.add(new_area)
                new_supervision.teachers.add(teacher)

                # Sync supervision
                new_supervision.sync()

                # Log
                if created:
                    logger.info("      New supervision created")

                # Save supervisions in reference dict
                if weekday not in supervisions_ref:
                    supervisions_ref[weekday] = {}
                if before_period_number not in supervisions_ref[weekday]:
                    supervisions_ref[weekday][before_period_number] = []
                supervisions_ref[weekday][before_period_number].append(new_supervision)

        for supervision in lesrooster_models.Supervision.objects.filter(
            rooms=new_area, break_slot__time_grid__validity_range=validity_range
        ):
            delete = True

            # Get weekday and period after break
            weekday = supervision.break_slot.weekday
            before_period_number = supervision.break_slot.extended_data["period_after"]

            # Delete supervision if no longer existing
            if (
                weekday in supervisions_ref
                and before_period_number in supervisions_ref[weekday]
                and supervision in supervisions_ref[weekday][before_period_number]
            ):
                delete = False

            if delete:
                with create_revision():
                    set_comment(_("Deleted by Untis import"))
                    supervision.delete()
                logger.info(f"    Supervision {supervision} deleted")

        ref[import_ref] = {"area": new_area, "supervisions": supervisions_ref}

    return ref


def import_slots(
    validity_range: ValidityRange,
) -> Tuple[
    Dict[int, Dict[int, lesrooster_models.Slot]], Dict[int, Dict[int, lesrooster_models.BreakSlot]]
]:
    """Import time periods an breaks."""
    times = (
        run_default_filter(validity_range, mysql_models.Commondata.objects, filter_term=False)
        .filter(id=30)
        .order_by("number")
    )

    times_ref = {}
    periods_ref = {}
    for time_ in tqdm(times, desc="Import times", **TQDM_DEFAULTS):
        period = time_.number

        # Extract time
        start_time = time(time_.fieldbyte1, time_.fieldbyte2)
        end_time = time(time_.fieldbyte3, time_.fieldbyte4)

        times_ref[period] = (start_time, end_time)
        periods_ref[period] = time_

    periods = (
        run_default_filter(validity_range, mysql_models.Commondata.objects, filter_term=False)
        .filter(id=CommonDataId.PERIOD.value)
        .order_by("number", "number1")
    )

    time_grid = lesrooster_models.TimeGrid.objects.get(validity_range=validity_range, group=None)

    slots_ref = {}
    breaks_ref = {}
    for time_period in tqdm(periods, desc="Import time slots", **TQDM_DEFAULTS):
        weekday = time_period.number - 1
        period = time_period.number1
        start_time = times_ref[period][0]
        end_time = times_ref[period][1]
        time_ = periods_ref[period]
        slot_name = time_.fieldtext50a if time_.fieldtext50a else ""

        logger.info(f"Import time slot on weekday {weekday} in the {period}. period")

        new_slot, created = lesrooster_models.Slot.objects.update_or_create(
            weekday=weekday,
            period=period,
            time_grid=time_grid,
            defaults={"time_start": start_time, "time_end": end_time, "name": slot_name},
        )

        if created:
            logger.info("  New time period created")

        # Create break
        break_name = time_.fieldtext50c if time_.fieldtext50c else ""
        break_end_time = start_time
        if period - 1 not in times_ref:
            break_start_time = (
                datetime.combine(datetime.today(), break_end_time) - timedelta(minutes=15)
            ).time()  # Workaround as Untis does not declare a time interval for the first break
        else:
            break_start_time = times_ref[period - 1][1]

        possible_breaks = lesrooster_models.BreakSlot.objects.filter(
            weekday=weekday, time_grid=time_grid, extended_data__period_after=period
        )
        if possible_breaks:
            new_break = possible_breaks[0]
        else:
            new_break = lesrooster_models.BreakSlot(
                weekday=weekday,
                time_grid=time_grid,
            )
        new_break.extended_data["period_after"] = period
        new_break.time_start = break_start_time
        new_break.time_end = break_end_time
        new_break.name = break_name
        new_break.save()

        # Build index with time periods
        if weekday not in slots_ref:
            slots_ref[weekday] = {}
        slots_ref[weekday][period] = new_slot

        # Build index with breaks
        if weekday not in breaks_ref:
            breaks_ref[weekday] = {}
        breaks_ref[weekday][period] = new_break

    return slots_ref, breaks_ref


def import_absence_reasons(
    validity_range: ValidityRange,
) -> Dict[int, kolego_models.AbsenceReason]:
    """Import absence reasons."""
    ref = {}

    # Get reasons
    reasons = run_default_filter(
        validity_range, mysql_models.Absencereason.objects, filter_term=False
    )

    for reason in tqdm(reasons, desc="Import absence reasons", **TQDM_DEFAULTS):
        if not reason.name:
            raise RuntimeError(
                "Absence reason ID {}: Cannot import absence reason without short name.".format(
                    reason.absence_reason_id
                )
            )

        # Build values
        short_name = reason.name
        name = reason.longname if reason.longname else short_name
        import_ref = reason.absence_reason_id

        logger.info(f"Import absence reason {short_name} …")

        new_reason, created = kolego_models.AbsenceReason.objects.update_or_create(
            short_name=short_name,
            defaults={"name": name},
        )

        if created:
            logger.info("  New absence reason created")

        new_reason.extended_data["import_ref_untis"] = import_ref
        new_reason.save()

        ref[import_ref] = new_reason

    return ref
