import logging

from django.utils.translation import gettext as _

from reversion import create_revision, set_comment
from tqdm import tqdm

from aleksis.apps.chronos.models import LessonEvent
from aleksis.apps.lesrooster.models import ValidityRange

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    connect_untis_fields,
    date_to_untis_date,
    get_first_period,
    get_last_period,
    move_weekday_to_range,
    run_default_filter,
    untis_date_to_date,
    untis_split_second,
)

logger = logging.getLogger(__name__)


def import_events(
    validity_range: ValidityRange,
    slots_ref,
    teachers_ref,
    classes_ref,
    rooms_ref,
):
    ref = {}

    # Get absences
    events = (
        run_default_filter(validity_range, mysql_models.Event.objects, filter_term=False)
        .filter(
            datefrom__lte=date_to_untis_date(validity_range.date_end),
            dateto__gte=date_to_untis_date(validity_range.date_start),
        )
        .order_by("event_id")
    )

    existing_events = []
    for event in tqdm(events, desc="Import events", **TQDM_DEFAULTS):
        import_ref = event.event_id

        logger.info(f"Import event {import_ref}")

        # Build values
        comment = event.text or ""

        datetime_start = untis_date_to_date(event.datefrom)
        datetime_end = untis_date_to_date(event.dateto)
        slot_start = event.lessonfrom
        slot_end = event.lessonto
        weekday_from = datetime_start.weekday()
        weekday_to = datetime_end.weekday()

        # Check min/max weekdays
        weekday_from = move_weekday_to_range(slots_ref, weekday_from)
        weekday_to = move_weekday_to_range(slots_ref, weekday_to)

        # Check min/max periods
        first_slot = get_first_period(slots_ref, weekday_from)
        last_slot = get_last_period(slots_ref, weekday_from)

        if slot_start == 0:
            slot_start = first_slot
        if slot_end == 0:
            slot_end = last_slot

        slot_start = slots_ref[weekday_from][slot_start]
        slot_end = slots_ref[weekday_to][slot_end]

        groups = []
        teachers = []
        rooms = []

        # Get groups, teachers and rooms
        raw_events = connect_untis_fields(event, "eventelement", 10)
        for raw_event in raw_events:
            el = untis_split_second(raw_event, remove_empty=False)

            # Group
            if el[0] != "0" and el[0] != "":
                group = classes_ref[int(el[0])]
                groups.append(group)

            # Teacher
            if el[2] != "0" and el[2] != "":
                teacher = teachers_ref[int(el[2])]
                teachers.append(teacher)

            # Room
            if el[3] != "0" and el[3] != "":
                room = rooms_ref[int(el[3])]
                rooms.append(room)

        datetime_start = slot_start.get_datetime_start(datetime_start)
        datetime_end = slot_end.get_datetime_end(datetime_end)

        new_event, created = LessonEvent.objects.update_or_create(
            extended_data__event_untis=import_ref,
            defaults={
                "datetime_start": datetime_start,
                "datetime_end": datetime_end,
                "title": comment,
                "slot_number_start": slot_start.period,
                "slot_number_end": slot_end.period,
            },
        )

        new_event.extended_data["event_untis"] = import_ref
        new_event.save()

        if created:
            logger.info("  New event created")

        # Sync m2m-fields
        new_event.groups.set(groups)
        new_event.teachers.set(teachers)
        new_event.rooms.set(rooms)

        existing_events.append(import_ref)
        ref[import_ref] = new_event

        # Delete all no longer existing events
        for e in LessonEvent.objects.filter(
            extended_data__event_untis__isnull=False,
            datetime_start__lte=validity_range.date_end,
            datetime_end__gte=validity_range.date_start,
        ):
            if e.extended_data["event_untis"] not in existing_events:
                logger.info(f"Event {e.id} deleted")
                with create_revision():
                    set_comment(_("Deleted by Untis import"))
                    e.delete()
