import logging

from django.db.models import Q
from django.utils.translation import gettext as _

import recurrence
import reversion
from reversion import set_comment
from tqdm import tqdm

from aleksis.apps.cursus import models as cursus_models
from aleksis.apps.lesrooster import models as lesrooster_models
from aleksis.core import models as core_models
from aleksis.core.util.core_helpers import get_site_preferences

from .... import models as mysql_models
from ....data_checks import CourseGroupNotFoundAndCreated, CourseGroupNotFoundAndNotCreated
from ..util import (
    TQDM_DEFAULTS,
    compare_m2m,
    connect_untis_fields,
    run_default_filter,
    untis_split_first,
    untis_split_second,
    untis_split_third,
)

logger = logging.getLogger(__name__)


def import_lessons(
    validity_range: lesrooster_models.ValidityRange,
    slots_ref,
    rooms_ref,
    subjects_ref,
    teachers_ref,
    classes_ref,
):
    """Import lessons."""
    # Lessons
    lessons = run_default_filter(validity_range, mysql_models.Lesson.objects)

    course_group_type = get_site_preferences()["untis_mysql__group_type_course_groups"]

    existing_lessons = []
    for lesson in tqdm(lessons, desc="Import lessons", **TQDM_DEFAULTS):
        lesson_id = lesson.lesson_id

        logger.info(_("Import lesson {}").format(lesson_id))

        if not lesson.lesson_tt:
            logger.warning(_("  Skip because missing times").format(lesson_id))
            continue

        existing_lessons.append(lesson_id)

        # Split data (,)
        raw_lesson_data = connect_untis_fields(lesson, "lessonelement", 10)
        raw_time_data = untis_split_first(lesson.lesson_tt, remove_empty=False)

        raw_time_data_2 = []
        for el in raw_time_data:
            # Split data (~)
            raw_time_data_2.append(untis_split_second(el, remove_empty=False))

        # Get time periods and rooms
        slots = []
        rooms_per_periods = []
        for el in raw_time_data_2:
            weekday = int(el[1]) - 1
            hour = int(el[2])
            room_ids = untis_split_third(el[3], conv=int)

            # Get rooms
            rooms = []
            for room_id in room_ids:
                r = rooms_ref[room_id]
                rooms.append(r)

            # Get time period
            slot = slots_ref[weekday][hour]
            slots.append(slot)
            rooms_per_periods.append(rooms)

        # Split data more (~)
        raw_lesson_data_2 = []
        for el in raw_lesson_data:
            raw_lesson_data_2.append(untis_split_second(el, remove_empty=False))

        use_room_idx = 0
        current_teacher_id = None
        # All part lessons (courses)
        for i, el in enumerate(raw_lesson_data_2):
            logger.info(f"  Lesson part {i}")

            # Get plain ids
            teacher_id = int(el[0]) if el[0] else 0
            subject_id = int(el[2])
            class_ids = untis_split_third(el[17], conv=int)

            # Get teacher
            teacher = teachers_ref[teacher_id] if teacher_id != 0 else None

            teachers = [teacher] if teacher else []

            # Get subject
            if subject_id != 0:
                subject = subjects_ref[subject_id]
            else:
                logger.warning(_("    Skip because missing subject"))
                continue

            if current_teacher_id and current_teacher_id != teacher_id:
                use_room_idx += 1

            # Get classes
            course_classes = []
            for class_id in class_ids:
                c = classes_ref[class_id]
                course_classes.append(c)

            logger.debug(
                f"    Teacher: {teacher} #{teacher_id}; Subject: {subject} #{subject_id}; "
                f"Classes: {course_classes} #{class_ids}"
            )

            course_group_not_found_and_not_created = False
            course_group = None
            if get_site_preferences()["untis_mysql__use_course_groups"]:
                # Negative import_ref denotes a course group
                group_import_ref = -int(f"{lesson_id}{i}")

                # Search by parent groups, teachers/owners and subject
                qs = (
                    core_models.Group.objects.filter(
                        parent_groups__in=[c.id for c in course_classes],
                        extended_data__subject_id=subject.id,
                    )
                    .filter(Q(school_term__isnull=True) | Q(school_term=validity_range.school_term))
                    .distinct()
                )

                if not qs.exists():
                    logger.warning("    No matching course group found")
                else:
                    logger.debug(f"{qs.count()} possibly matching course groups found: {qs}")

                # Check if found groups match
                possible_groups = []
                for found_group in qs:
                    if compare_m2m(course_classes, found_group.parent_groups.all()):
                        possible_groups.append(found_group)

                logger.debug(
                    f"    {len(possible_groups)} possible groups found "
                    f"by searching by parent groups: {possible_groups}"
                )

                if len(possible_groups) == 1:
                    course_group = possible_groups[0]
                    logger.info(
                        "    Course group found by searching "
                        f"by parent groups, and subject: {course_group}"
                    )
                else:
                    for found_group in possible_groups:
                        if compare_m2m(teachers, found_group.owners.all()):
                            course_group = found_group
                            logger.info(
                                "    Course group found by searching by parent groups, "
                                "teachers (owners) and subject"
                            )

                    if not course_group:
                        logger.debug(
                            "    No course group found by searching "
                            "by parent groups, teachers (owners) and subject"
                        )

                if (
                    not course_group
                    and get_site_preferences()["untis_mysql__course_groups_fuzzy_matching"]
                ):
                    logger.debug("    Fuzzy matching mode used")
                    qs = qs.filter(owners__in=[t.id for t in teachers])
                    if qs.count() != 1:
                        logger.warning(
                            "    Course group not found or more than one found "
                            f"({qs.count()} groups found: {qs}) by searching by parent groups, "
                            "teachers (owners) and subject (fuzzy matching mode)"
                        )
                    else:
                        course_group = qs.first()
                        logger.info(
                            "    Course group found by searching by parent groups, "
                            f"teachers (owners) and subject (fuzzy matching mode): {course_group}"
                        )

                changed = False
                register_data_check = get_site_preferences()[
                    "untis_mysql__data_check_for_not_found_course_groups"
                ]
                if not course_group and get_site_preferences()["untis_mysql__create_course_groups"]:
                    # No matching group found
                    logger.info("    No matching course group found, generate one")

                    # Build names and refs for course groups
                    group_short_name = "{}-{}".format(
                        "".join([c.short_name for c in course_classes]),
                        subject.short_name,
                    )
                    group_name = "{}: {}".format(
                        ", ".join([c.short_name for c in course_classes]),
                        subject.short_name,
                    )

                    # Get or create course group
                    course_group, created = core_models.Group.objects.get_or_create(
                        short_name=group_short_name, defaults={"name": group_name}
                    )

                    # Log
                    if created:
                        logger.info("    Course group created")
                    else:
                        logger.info("    Existing course group found")

                    # Update parent groups
                    course_group.parent_groups.set(course_classes)

                    # Update name
                    if course_group.name != group_name:
                        course_group.name = group_name
                        logger.info("    Name of course group updated")

                        changed = True

                    if created and register_data_check:
                        # No match, new course group created, so create a data problem if activated
                        CourseGroupNotFoundAndCreated.register_result(course_group)
                elif register_data_check:
                    # No match, create deactivated, so create a data problem later if activated
                    course_group_not_found_and_not_created = True

                if course_group:
                    # Update owners
                    course_group.owners.set(teachers)

                    # Update import ref
                    if course_group.extended_data.get("import_ref_untis") != group_import_ref:
                        course_group.extended_data["import_ref_untis"] = group_import_ref
                        logger.info("    Import reference of course group updated")
                        changed = True

                    if course_group.extended_data.get("subject_id") != subject.id:
                        course_group.extended_data["subject_id"] = subject.id
                        logger.info("    Subject reference of course group updated")
                        changed = True

                    if course_group.school_term != validity_range.school_term:
                        course_group.school_term = validity_range.school_term
                        logger.info("    School term reference of course group updated")
                        changed = True

                    if course_group_type and course_group.group_type != course_group_type:
                        course_group.group_type = course_group_type
                        logger.info("    Group type of course group updated")
                        changed = True

                    if changed:
                        course_group.save()

                    groups = [course_group]

                else:
                    groups = []
            else:
                groups = course_classes

            # Get course
            possible_courses = cursus_models.Course.objects.filter(
                subject=subject,
                groups__in=groups,
            )
            course = None
            for possible_course in possible_courses:
                if compare_m2m(possible_course.groups.all(), groups):
                    course = possible_course
                    break

            if not course:
                if course_group:
                    name = course_group.name
                else:
                    name = "{}: {}".format(
                        ", ".join([c.short_name or c.name for c in groups]),
                        subject.short_name or subject.name,
                    )
                course = cursus_models.Course.objects.create(subject=subject, name=name)
            course.groups.set(groups or [])
            course.teachers.set(teachers)

            # All times for this course
            old_lessons_qs = lesrooster_models.Lesson.objects.filter(
                extended_data__lesson_id_untis=lesson_id,
                extended_data__element_id_untis=i,
                slot_start__time_grid__validity_range=validity_range,
            )

            # Each lesson time is exactly one slot
            slots_count = old_lessons_qs.count()

            (
                timebound_course_config,
                __,
            ) = lesrooster_models.TimeboundCourseConfig.objects.update_or_create(
                course=course,
                validity_range=validity_range,
                defaults=dict(lesson_quota=slots_count),
            )
            timebound_course_config.teachers.set(teachers)

            existing_lesson_pks = []

            # Sync time periods
            for j, slot in enumerate(slots):
                logger.info(f"    Import lesson {slot}")

                # Get room if provided
                rooms = rooms_per_periods[j]
                room = rooms[use_room_idx] if use_room_idx < len(rooms) else None

                # Check if an old lesson is provided
                old_lesson_qs = old_lessons_qs.filter(slot_start=slot)
                if old_lesson_qs.exists():
                    # Update old lesson

                    lesson = old_lesson_qs[0]
                    lesson.course = course
                    lesson.subject = subject
                    lesson.slot_end = slot
                    logger.info("      Course, subject, and slot updated")
                else:
                    # Create new lesson

                    lesson = lesrooster_models.Lesson.objects.create(
                        course=course,
                        slot_start=slot,
                        slot_end=slot,
                        subject=subject,
                    )
                    logger.info("      New lesson added")

                lesson.recurrence = lesson.build_recurrence(recurrence.WEEKLY)

                lesson.extended_data["lesson_id_untis"] = lesson_id
                lesson.extended_data["element_id_untis"] = i

                lesson.save()

                lesson.rooms.set([room] if room else [])
                lesson.teachers.set(teachers)

                # Sync lesson to lesson event
                lesson.sync()

                if course_group_not_found_and_not_created:
                    # Register data problem here because before the lesson object is unknown
                    CourseGroupNotFoundAndNotCreated.register_result(lesson)

                existing_lesson_pks.append(lesson.pk)

            # delete all no-longer existing lesson periods
            old_lessons_qs.exclude(pk__in=existing_lesson_pks).delete()
            logger.info("    Old lessons deleted")
            current_teacher_id = teacher_id

    for lesson in lesrooster_models.Lesson.objects.filter(
        slot_start__time_grid__validity_range=validity_range,
        extended_data__lesson_id_untis__isnull=False,
    ):
        if lesson.extended_data["lesson_id_untis"] not in existing_lessons:
            logger.info(f"Lesson {lesson.id} deleted")
            with reversion.create_revision():
                set_comment(_("Deleted by Untis import"))
                lesson.delete()
