import logging
from enum import Enum

from django.db.models import Q

from tqdm import tqdm

from aleksis.apps.chronos.models import LessonEvent, SupervisionEvent
from aleksis.apps.lesrooster import models as lesrooster_models

from .... import models as mysql_models
from ..util import (
    TQDM_DEFAULTS,
    date_to_untis_date,
    run_default_filter,
    untis_date_to_date,
    untis_split_first,
)

logger = logging.getLogger(__name__)


class SubstitutionFlag(Enum):
    CANCELLED = "E"
    CANCELLED_FOR_TEACHERS = "F"


def import_substitutions(
    validity_range: lesrooster_models.ValidityRange,
    teachers_ref,
    subjects_ref,
    rooms_ref,
    classes_ref,
    supervision_areas_ref,
    slots_ref,
    absence_substitutions=None,
):
    """Import substitutions."""
    if not absence_substitutions:
        absence_substitutions = []
    absence_substitutions = [sub.absence_ref_untis for sub in absence_substitutions]

    subs = (
        run_default_filter(validity_range, mysql_models.Substitution.objects, filter_term=False)
        .filter(
            date__gte=date_to_untis_date(validity_range.date_start),
            date__lte=date_to_untis_date(validity_range.date_end),
        )
        .exclude(
            Q(flags__contains="N")
            | Q(flags__contains="b")
            | Q(flags__contains="F")
            | Q(flags__exact="g")
        )
        .order_by("substitution_id")
    )

    existing_subs = []
    for sub in tqdm(subs, desc="Import substitutions", **TQDM_DEFAULTS):
        # IDs
        sub_id = sub.substitution_id
        existing_subs.append(sub_id)

        lesson_id = sub.lesson_idsubst

        logger.info(f"Import substitution {sub_id}")

        # Time
        date = untis_date_to_date(sub.date)
        weekday = date.weekday()
        period = sub.lesson

        # Supervision substitution?
        is_supervision_substitution = sub.corridor_id != 0

        # Cancellation?
        cancelled, cancelled_for_teachers = False, False
        if SubstitutionFlag.CANCELLED.value in sub.flags:
            cancelled = True
        elif SubstitutionFlag.CANCELLED_FOR_TEACHERS.value in sub.flags:
            cancelled_for_teachers = True

        # Comment
        comment = sub.text if sub.text else ""

        # Teacher
        teacher_old = teachers_ref[sub.teacher_idlessn] if sub.teacher_idlessn != 0 else None

        teachers = []
        teacher_new = None
        if sub.teacher_idsubst != 0:
            teacher_new = teachers_ref[sub.teacher_idsubst]
            teachers = [teacher_new]

            if teacher_old is not None and teacher_new.id == teacher_old.id:
                teachers = []

        if not is_supervision_substitution:
            lessons = lesrooster_models.Lesson.objects.filter(
                slot_start__time_grid__validity_range=validity_range,
                extended_data__lesson_id_untis=lesson_id,
                teachers=teacher_old,
                slot_start__period=period,
                slot_start__weekday=weekday,
            )

            if lessons.exists():
                lesson = lessons[0]
                logger.info(f"  Matching lesson found ({lesson})")
            else:
                lesson = None

            # Subject
            subject_old = lesson.subject if lesson else None
            subject_new = None
            if sub.subject_idsubst != 0:
                subject_new = subjects_ref[sub.subject_idsubst]

                if subject_old and subject_old.id == subject_new.id:
                    subject_new = None

            if cancelled:
                subject_new = None

            # Room
            rooms_old = [room for room in lesson.rooms.all()] if lesson else []
            rooms_old_ids = [room.id for room in rooms_old]
            rooms = []
            room_new = None
            if sub.room_idsubst != 0:
                room_new = rooms_ref[sub.room_idsubst]
                rooms = [room_new]

                if rooms_old is not None and room_new.id in rooms_old_ids:
                    rooms = []

            # Classes
            classes = []
            class_ids = untis_split_first(sub.classids, conv=int)

            for id_ in class_ids:
                classes.append(classes_ref[id_])

            # Skip incomplete substitutions
            if (
                not rooms
                and not subject_new
                and not teachers
                and not cancelled
                and not cancelled_for_teachers
                and not comment
            ):
                continue

            if lesson:
                datetime_start = lesson.slot_start.get_datetime_start(date)
                datetime_end = lesson.slot_end.get_datetime_end(date)

                (
                    substitution,
                    created,
                ) = (
                    LessonEvent.objects.select_related(None)
                    .prefetch_related(None)
                    .update_or_create(
                        amends=lesson.lesson_event,
                        extended_data__substitution_ref_untis=sub_id,
                        defaults={
                            "datetime_start": datetime_start,
                            "datetime_end": datetime_end,
                            "subject": subject_new,
                            "cancelled": cancelled,
                            "comment": comment,
                            "slot_number_start": lesson.slot_start.period,
                            "slot_number_end": lesson.slot_end.period,
                        },
                    )
                )

                if created:
                    logger.info("  Substitution created")

                # Sync teachers
                substitution.teachers.set(teachers)
                substitution.rooms.set(rooms)
                logger.info("   Substitution teachers and rooms set")

                # Update import ref
                substitution.extended_data["substitution_ref_untis"] = sub_id
                substitution.save()

            else:
                logger.info("  Extra lesson detected")
                slot = slots_ref[date.weekday()][period]

                groups = [classes_ref[pk] for pk in untis_split_first(sub.classids, int)]

                subject = subject_new if subject_new else subject_old
                if not teacher_new and not teacher_old:
                    teachers = []
                else:
                    teachers = [teacher_old] if not teacher_new else [teacher_new]

                if not subject:
                    logger.warning("  Skip because missing subject")
                    continue

                datetime_start = slot.get_datetime_start(date)
                datetime_end = slot.get_datetime_end(date)

                (
                    extra_lesson,
                    created,
                ) = (
                    LessonEvent.objects.select_related(None)
                    .prefetch_related(None)
                    .update_or_create(
                        extended_data__extra_lesson_untis=sub_id,
                        defaults={
                            "datetime_start": datetime_start,
                            "datetime_end": datetime_end,
                            "subject": subject,
                            "comment": comment,
                            "slot_number_start": slot.period,
                            "slot_number_end": slot.period,
                        },
                    )
                )

                extra_lesson.extended_data["extra_lesson_untis"] = sub_id
                extra_lesson.save()

                if created:
                    logger.info("  Extra lesson created")

                extra_lesson.rooms.set(rooms)
                extra_lesson.teachers.set(teachers)
                extra_lesson.groups.set(groups)
        else:
            if teacher_new:
                logger.info("  Supervision substitution detected")

                # Supervision
                area_ref = supervision_areas_ref[sub.corridor_id]
                possible_supervisions = area_ref["supervisions"][weekday][period]

                supervision = None
                for possible_supervision in possible_supervisions:
                    if teacher_old in possible_supervision.teachers.all():
                        supervision = possible_supervision

                logger.info(f"  Supervision {supervision} is replaced by {teacher_new}")

                if supervision:
                    datetime_start = supervision.break_slot.get_datetime_start(date)
                    datetime_end = supervision.break_slot.get_datetime_end(date)

                    (
                        substitution,
                        created,
                    ) = SupervisionEvent.objects.update_or_create(
                        amends=supervision.supervision_event,
                        extended_data__supervision_substitution_ref_untis=sub_id,
                        defaults={"datetime_start": datetime_start, "datetime_end": datetime_end},
                    )

                    substitution.teachers.set(teachers)

                    substitution.extended_data["supervision_substitution_ref_untis"] = sub_id
                    substitution.save()

                    if created:
                        logger.info("  Supervision substitution created")

    # Delete all no longer existing substitutions
    lesson_substitutions_in_range = LessonEvent.objects.filter(
        datetime_end__date__gte=validity_range.date_start,
        datetime_start__date__lte=validity_range.date_end,
        amends__isnull=False,
    )
    for s in lesson_substitutions_in_range:
        if (
            s.extended_data.get("substitution_ref_untis")
            and s.extended_data["substitution_ref_untis"] not in existing_subs
        ) or (
            s.extended_data.get("absence_ref_untis")
            and s.extende_data["absence_ref_untis"] not in absence_substitutions
        ):
            logger.info(f"Substitution {s.id} deleted")
            s.delete()

    lesson_events = lesrooster_models.Lesson.objects.filter(
        slot_start__time_grid__validity_range=validity_range
    ).values_list("lesson_event__pk", flat=True)

    lesson_substitutions_in_range.exclude(amends__pk__in=lesson_events).delete()

    # Delete all no longer existing extra lessons
    for s in LessonEvent.objects.filter(
        datetime_end__date__gte=validity_range.date_start,
        datetime_start__date__lte=validity_range.date_end,
        extended_data__extra_lesson_untis__isnull=False,
    ):
        if s.extended_data["extra_lesson_untis"] not in existing_subs:
            logger.info(f"Extra lesson {s.id} deleted")
            s.delete()

    # Delete all no longer existing supervision substitutions
    supervision_substitutions_in_range = LessonEvent.objects.filter(
        datetime_end__date__gte=validity_range.date_start,
        datetime_start__lte=validity_range.date_end,
        extended_data__supervision_substitution_ref_untis__isnull=False,
        amends__isnull=False,
    )
    for s in supervision_substitutions_in_range:
        if s.extended_data["supervision_substitution_ref_untis"] not in existing_subs:
            logger.info(f"Supervision substitution {s.id} deleted")
            s.delete()

    supervision_events = lesrooster_models.Supervision.objects.filter(
        break_slot__time_grid__validity_range=validity_range
    ).values_list("supervision_event__pk", flat=True)

    supervision_substitutions_in_range.exclude(amends__pk__in=supervision_events).delete()
