from typing import Optional

from django.db.models import QuerySet

import reversion
from tqdm import tqdm

from .importers.absences import import_absences
from .importers.common_data import (
    import_absence_reasons,
    import_classes,
    import_rooms,
    import_slots,
    import_subjects,
    import_supervision_areas,
    import_teachers,
)
from .importers.events import import_events
from .importers.holidays import import_holidays
from .importers.lessons import import_lessons

# from aleksis.apps.chronos.util.change_tracker import TimetableDataChangeTracker
from .importers.substitutions import import_substitutions
from .importers.terms import import_terms
from .util import TQDM_DEFAULTS


def untis_import_mysql(
    terms: Optional[QuerySet] = None, school_id: Optional[int] = None, version: Optional[int] = None
):
    # School terms and validity ranges
    validity_ref = import_terms(terms, school_id=school_id, version=version)

    for validity_range in tqdm(
        validity_ref.values(), desc="Import data for terms", **TQDM_DEFAULTS
    ):
        with reversion.create_revision(atomic=True):
            reversion.set_comment(f"Untis import for validity range {validity_range}")

            # tracker = TimetableDataChangeTracker()

            # Common data for Chronos
            subjects_ref = import_subjects(validity_range)
            rooms_ref = import_rooms(validity_range)
            absence_reasons_ref = import_absence_reasons(validity_range)

            # Common data for core
            teachers_ref = import_teachers(validity_range)
            classes_ref = import_classes(validity_range, teachers_ref)

            # Slots and breaks
            slots_ref, breaks_ref = import_slots(validity_range)

            # Holidays
            holidays_ref = import_holidays(validity_range)  # noqa:  F841

            # Supervisions
            supervision_areas_ref = import_supervision_areas(
                validity_range, breaks_ref, teachers_ref
            )

            # Lessons
            import_lessons(
                validity_range,
                slots_ref,
                rooms_ref,
                subjects_ref,
                teachers_ref,
                classes_ref,
            )

            # Substitutions
            __, created_substitutions = import_absences(
                validity_range,
                absence_reasons_ref,
                slots_ref,
                teachers_ref,
                classes_ref,
                rooms_ref,
            )
            import_substitutions(
                validity_range,
                teachers_ref,
                subjects_ref,
                rooms_ref,
                classes_ref,
                supervision_areas_ref,
                slots_ref,
                created_substitutions,
            )

            # Events
            import_events(validity_range, slots_ref, teachers_ref, classes_ref, rooms_ref)
